/* Copyright (c) 2001 Mark Pruett, Ben Woodard, James Antill
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version 2
 * of the Licence, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRENTY; without even the implied warrenty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public Licence in the COPYING file for more
 * details.
 *
 * You should have received a copy of the GNU Library General 
 * Public License along with the GNU C Library; see the file 
 * COPYING.LIB.  If not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#ifndef __gpc_pdinfo_h__
#define __gpc_pdinfo_h__
enum { PDDB_LOCAL, PDDB_REMOTE};

typedef struct {
    int x;
    int y;
    char *comment;
} GPC_PDINFO_RES;

typedef struct {
    char *bitsperpixel;
    char *comment;
} GPC_PDINFO_BITSPERPIXEL;

typedef struct {
    char **printer_name;	
    char *printer_manuf;
    char *gsdriver;
    char *brief_desc;
    char *desc;
    char *ppdfile;	
    GPC_PDINFO_RES **res_list;
    GPC_PDINFO_BITSPERPIXEL **bits_list;	
} GPC_PDINFO;


/*----------------------------------------------------------------
  Parse the xml file containing pddb information.

  Given a file name of a local PDDB file, this routine parses the
  file and produces a list of GPC_PDINFO structures containing
  the file's contents.

  Returns a null-terminated array of PDDB records, or NULL if 
  an error occurs.
-------------------------------------------------------------------*/
extern GPC_PDINFO ** gpc_pdinfo_get_list(const char *filename,char **url_list);
/* filename -- the name of the PDDB (XML formatted) file */
/* url_list -- reserved for future use. */

/*----------------------------------------------------------------
  Free a previously-allocated GPC_PDINFO list.

  Returns NULL.
----------------------------------------------------------------*/
extern GPC_PDINFO ** gpc_pdinfo_free_list(GPC_PDINFO **list);

/*----------------------------------------------------------------
  Insert an item into the internal PDDB list.

  Given a list and a new item, insert the new item into the list
  in sorted (by printer_name) order.

  Returns the new list.
----------------------------------------------------------------*/
extern GPC_PDINFO ** gpc_pdinfo_insert (GPC_PDINFO ** list,GPC_PDINFO *item);
/* list -- the list of PDDB information structures */
/* item -- the PDDB record to insert */
/*----------------------------------------------------------------
  Find a printer in the printer list.

  Given the list and the name of a printer, search the list and 
  find an item whose printer_name field matches.

  Returns a pointer to the matching list item if found, or 
  NULL otherwise.
----------------------------------------------------------------*/
extern GPC_PDINFO *gpc_pdinfo_find(GPC_PDINFO **list,
                                   const char *printer_name);

/* list -- the list of PDDB information structures */ 
/* printer_name -- the exact name of the printer */

extern GPC_PDINFO *gpc_pdinfo_description_find(GPC_PDINFO **list,
                                               const char *printer_desc);

/*----------------------------------------------------------------
  Save the GPC_PDINFO list to an local PDDB XML file.

  Returns 0 on success, non-zero otherwise;
----------------------------------------------------------------*/
extern int gpc_pdinfo_save_local(GPC_PDINFO **list,const char *filename);
/* list -- list of Printer Driver DB information */
/* filename -- XML file where data will be saved */

/*----------------------------------------------------------------
  Get a list of unique manufacturers names.

  Given a list of PDDB records, build a list of unique 
  manufacturers names, extracted from the printer_manuf fields.

  Returns a sorted, null-terminated list of manufacturer's names.
----------------------------------------------------------------*/
extern char ** gpc_pdinfo_get_manuf_names (GPC_PDINFO **list);

/*----------------------------------------------------------------
  Frees a list of manufacturer's names

  Free a list of manufacturer's names previously allocated by
  gpc_pdinfo_get_manuf_names().
----------------------------------------------------------------*/
extern void gpc_pdinfo_free_manuf_names(char **list);

/*----------------------------------------------------------------
  Get a list of PDDB records for a single manufacturer.

  Given the list of all PDDB records and the name of a target 
  manufacturer, return a list of pointers to matching records
  within the list.
----------------------------------------------------------------*/
extern GPC_PDINFO ** gpc_pdinfo_get_manuf_list(GPC_PDINFO **list,char *manuf);
/* list -- list of Printer Driver DB records */
/* manuf -- the target manufacturer name */

/*----------------------------------------------------------------
  Free a list of pointers to PDDB records for a specific 
  manufacturer. This routine doesn't delete the records 
  themselves, as they are still part of the complete list of PDDB 
  records allocated with gpc_pdinfo_get_list().
----------------------------------------------------------------*/
/* list of PDDB records for one manufacturer */
extern void gpc_pdinfo_free_manuf_list(GPC_PDINFO **manlist);
extern int gpc_pdinfo_set_manufacturer(char * manuf);
extern int gpc_pdinfo_get_dbtype(void);

#endif
