#include <stdlib.h>

#include <gnome.h>

#include "gpc_pdinfo.h"
#include "gpc_support.h"
#include <assert.h>
#include "gpc_queue.h"
#include "../libprintsys/src/printsys.h"

static GPC_QUEUE_LIST **q_ret = NULL;
static const char *printcap_filename;

void gpc__queue_free_filter_opts(GPC_FILTER_SELECTION *filter_opts)
{
  if (!filter_opts)
    return;
  
  g_free(filter_opts->res);
  g_free(filter_opts->size);
  g_free(filter_opts->bpp);
  g_free(filter_opts->margin_right_left);
  g_free(filter_opts->margin_top_bottom);
  g_free(filter_opts->pages_per_page);
  g_free(filter_opts->gs_opts);
  g_free(filter_opts);
}

static void gpc__queue_free_info(GPC_QUEUE_INFO *qinfo)
{
  unsigned int scan = 0;

  while (qinfo->qlist[scan])
  {
    free(qinfo->qlist[scan]); /* FIXME: arghh different */
    ++scan;
  }
  
  gpc__queue_free_filter_opts(qinfo->filter_opts);
  
  g_free(qinfo->qlist);
  g_free(qinfo->spool_dir);
  g_free(qinfo);
}

static void del_old_names(GPC_QUEUE_INFO *qinfo, LPS_Printer_t *prn)
{
  wchar_t **cur = NULL;
  char **curname = NULL;
  wchar_t *wtmp = NULL;
  
  // remove the names which are no longer relevant
  for(cur=prn->names;*cur!=NULL;cur++){
    int matched=0;
    for(curname=qinfo->qlist;*curname!=NULL;curname++){
      wtmp=lps_promote(*curname);
      if(!wcscmp(wtmp,*cur)){
        free(wtmp);
	matched = 1;
	break;
      }
      free(wtmp);
    }
    
    if(matched==0){
      wchar_t **idx = cur;
      while (*idx)
        idx++;
      // squish the array
      free(*cur);
      memmove(cur, cur + 1, (idx - cur) * sizeof(wchar_t **));
      cur--; 
    }
  }
}

static void gpc__queue_unref(GPC_QUEUE_INFO *qinfo)
{
  if (!--qinfo->ref)
    gpc__queue_free_info(qinfo);
}

static void gpc__queue_ref(GPC_QUEUE_INFO *qinfo)
{
  ++qinfo->ref;
}

static char *get_str(LPS_Printer_t *prn, const wchar_t *key){
  char *retval;
  const wchar_t *wtmp=lps_pr_lookup_field(prn,key);
  if(wtmp){
    char *tmp=lps_demote(wtmp);
    assert(tmp);
    retval=g_strdup(tmp);
  }else
    retval=NULL;
  return retval;
}

static unsigned int get_bool(LPS_Printer_t *prn, const wchar_t *key)
{
  const wchar_t *wtmp = lps_pr_lookup_field(prn, key);
  
  if (wtmp)
    return (TRUE);

  return (FALSE);
}

static size_t gpc__queue_len(GPC_QUEUE_LIST **qlist)
{
  size_t scan = 0;

  if (!qlist)
    return (0);
  
  while (qlist[scan])
    ++scan;

  return (scan);
}

static size_t gpc__queue_node_len(GPC_QUEUE_INFO *qinfo)
{
  size_t scan = 0;

  if (!qinfo || !qinfo->qlist)
    return (0);
  
  while (qinfo->qlist[scan])
    ++scan;

  return (scan);
}

static GPC_QUEUE_LIST **gpc__queue_qlist_add_node(GPC_QUEUE_LIST **qlist,
                                                  const size_t cur_len,
                                                  const char *qname,
                                                  GPC_QUEUE_INFO *qinfo)
{
  qlist = g_realloc(qlist, (cur_len + 2) * sizeof(GPC_QUEUE_LIST *));
  q_ret = qlist; /* FIXME: */
  
  qlist[cur_len] = g_malloc(sizeof(GPC_QUEUE_LIST));
  qlist[cur_len]->qname = strdup(qname);
  qlist[cur_len]->qinfo = qinfo;
  gpc__queue_ref(qinfo);
  
  qlist[cur_len + 1] = NULL;

  return (qlist);
}

/* add the list ptrs in, from the vectors */
static void gpc__queue_add(GPC_QUEUE_LIST **qlist, GPC_QUEUE_INFO *qinfo)
{
  size_t len = gpc__queue_len(qlist);
  unsigned int scan = 0;
  
  while (qinfo->qlist[scan])
  {
    qlist = gpc__queue_qlist_add_node(qlist, len, qinfo->qlist[scan], qinfo);
    
    ++len;
    ++scan;
  }
}

GPC_QUEUE_LIST **gpc_queue_get_list(char *filename)
{
  LPS_System_t *sys;
  const wchar_t **printers;
  const wchar_t **cur;
  
  if(q_ret)
    return q_ret;

  q_ret=g_malloc(sizeof(GPC_QUEUE_LIST*));
  
  *q_ret=NULL;
  assert(q_ret);

  printcap_filename=(filename==NULL)?"/etc/printcap":filename;
  sys=lps_init(printcap_filename);
  assert(sys);
  printers=lps_get_printer_list(sys,NULL);
  assert(printers);

  for (cur=printers;*cur!=NULL;cur++)
  {
    LPS_Printer_t *prn;
    GPC_QUEUE_INFO *qinfo;
    char *tmp;
    wchar_t *wtmp;
    int ps_printer = FALSE;
    
    prn=lps_get_printer(sys,*cur,NULL);
    assert(prn);

    qinfo = g_malloc0(sizeof(GPC_QUEUE_INFO));
    qinfo->filter_opts = g_malloc0(sizeof(GPC_FILTER_SELECTION));
    
    // XXX note this is allocated using a regular malloc not a
    // g_malloc might cause problems somewhere.
    qinfo->qlist = lps_demote_vec((const wchar_t **)prn->names);

    gpc__queue_add(q_ret, qinfo);

    /* Do filter stuff first, before the continue statments ... */
    qinfo->filter_opts->printer_model = get_str(prn, L"PRINTER_MODEL");
    qinfo->filter_opts->gs_dev = get_str(prn, L"GSDEVICE");
    if (qinfo->filter_opts->gs_dev &&
        !strcmp(qinfo->filter_opts->gs_dev, "POSTSCRIPT"))
      ps_printer = TRUE;
    else
      ps_printer = FALSE;
    qinfo->filter_opts->res = get_str(prn, L"RESOLUTION");
    qinfo->filter_opts->size = get_str(prn, L"PAPERSIZE");
    qinfo->filter_opts->bpp = get_str(prn, L"BITSPERPIXEL");
    qinfo->filter_opts->margin_right_left = get_str(prn, L"RTLFTMAR");
    qinfo->filter_opts->margin_top_bottom = get_str(prn, L"TOPBOTMAR");
    qinfo->filter_opts->pages_per_page = get_str(prn, L"NUP");
    qinfo->filter_opts->send_eof = get_bool(prn, ps_printer ?
                                            L"TEXT_SEND_EOF" :
                                            L"PS_SEND_EOF");
    qinfo->filter_opts->fix_stair_text = get_bool(prn, L"CRLFTRANS");
    if (0)
    qinfo->filter_opts->fast_text = get_bool(prn, L""); /* ? */
    qinfo->filter_opts->gs_opts = get_str(prn, L"EXTRA_GS_OPTIONS");

    /* Do normal part of qinfo struct... */
    qinfo->spool_dir=get_str(prn,L"sd");

    qinfo->suppress_headers = get_bool(prn, L"sh");
    
    wtmp=lps_pr_lookup_field(prn,L"mx");
    if(wtmp){
      tmp=lps_demote(wtmp);
      assert(tmp);
      qinfo->file_limit=strtoul(tmp,NULL,10);
      free(tmp);
    }else
      qinfo->file_limit=0;
    
    qinfo->ifilter=get_str(prn,L"if"); // input filter
    qinfo->ofilter=get_str(prn,L"of"); // output filter
    qinfo->acct_file=get_str(prn,L"af"); // accounting file
    qinfo->local_device=get_str(prn,L"lp"); // device
    if(qinfo->local_device){
      qinfo->qtype=GPC_QTYPE_LOCAL;
      continue;
    }
    qinfo->remote_host=get_str(prn,L"rm"); // remote machine
    qinfo->remote_queue=get_str(prn,L"rp"); // remote queue
    if(qinfo->remote_host){
      qinfo->qtype=GPC_QTYPE_REMOTE;
      continue;
    }
    qinfo->direct_host=get_str(prn,L"printer_ip");
    wtmp=lps_pr_lookup_field(prn,L"port");
    if(wtmp){
      tmp=lps_demote(wtmp);
      assert(tmp);
      qinfo->direct_port=strtoul(tmp,NULL,10);
      free(tmp);
    }else
      qinfo->file_limit=0;
    if(qinfo->direct_host){
      qinfo->qtype=GPC_QTYPE_DIRECT;
      continue;
    }

    /* Right now this information is still stored in a .config
       file in the spooldir rather than in the printcap. We 
       need to put in code to read out of that source of 
       information */
    qinfo->SMB_ip=get_str(prn,L"smb_hostip");
    qinfo->SMB_host=get_str(prn,L"smb_host");
    qinfo->SMB_workgroup=get_str(prn,L"smb_workgroup");
    wtmp=lps_pr_lookup_field(prn,L"smb_share");
    if(wtmp){
      char *tmp2;
      tmp=lps_demote(wtmp);
      assert(tmp);
      if (strlen(tmp) > 2)
      {
        if ((tmp2=strchr(tmp+2,'/')))
        {
          *tmp2=0;
          qinfo->SMB_host=g_strdup(tmp+2);
          qinfo->SMB_printer_name=g_strdup(tmp2+1);
        }
      }
      free(tmp);
    }else
      qinfo->SMB_host=qinfo->SMB_printer_name=NULL;
    qinfo->user=get_str(prn,L"user");
    qinfo->passwd=get_str(prn,L"password");
    if(qinfo->SMB_ip){
      qinfo->qtype=GPC_QTYPE_SMB;
      continue;
    }
  
    // currently stored in .config file rather than in printcap
    qinfo->NCP_printserver_name=get_str(prn,L"ncp_server");
    qinfo->NCP_queue_name=get_str(prn,L"ncp_printer");
    if(qinfo->NCP_printserver_name)
      qinfo->qtype=GPC_QTYPE_NCP;
    else
      qinfo->qtype=GPC_QTYPE_UNKNOWN;
  }

  lps_end(sys);

  /* FIXME: doesn't duplicate */
  return (q_ret);
}

void gpc_queue_free_list (GPC_QUEUE_LIST ** qlist __attribute__ ((unused)))
{
  /* FIXME: can't alter atm. so no free */
}


GPC_QUEUE_INFO *gpc_queue_get_info(char *queue_name)
{
  unsigned int scan = 0;

  while (q_ret[scan])
  {
    if (!strcmp(q_ret[scan]->qname, queue_name))
      /* FIXME: doesn't duplicate */
      return (q_ret[scan]->qinfo);
    ++scan;
  }

  return (NULL);
}

void gpc_queue_free_info (GPC_QUEUE_INFO *qinfo __attribute__ ((unused)))
{
   /* FIXME: can't alter atm. so no free */
}

GPC_QUEUE_LIST **gpc_queue_del(GPC_QUEUE_LIST **qlist, char *queue_name)
{
  unsigned int scan = 0;
  LPS_System_t *sys=lps_init(printcap_filename);
  LPS_Printer_t *printer;
  wchar_t *wtmp=lps_promote(queue_name);

  assert(sys);
  assert(wtmp);
  printer=lps_get_printer(sys,wtmp,NULL);
  free(wtmp);
  assert(printer);
  
  while (qlist[scan]){
    if (!strcmp(qlist[scan]->qname, queue_name)){
      size_t len = gpc__queue_len(qlist);
      GPC_QUEUE_LIST *die = qlist[scan];
      
      memmove(qlist + scan, qlist + scan + 1,
              (len - scan) * sizeof(GPC_QUEUE_LIST *));

      scan = 0;
      while (die->qinfo->qlist[scan]){
        if (!strcmp(die->qinfo->qlist[scan], die->qname)){
          if (!scan && !die->qinfo->qlist[1]){
            lps_destroy_printer(printer, NULL);
            break;
          }

          free(die->qinfo->qlist[scan]);

          /* compress array */
          while (die->qinfo->qlist[scan]){
            die->qinfo->qlist[scan] = die->qinfo->qlist[scan + 1];
            ++scan;
          }
          
          die->qinfo->qlist = g_realloc(die->qinfo->qlist,
                                        sizeof(char *) * (scan + 1));
          del_old_names(die->qinfo, printer);
          lps_commit_printer(printer, NULL);
          continue;
        }
        
        ++scan;
      }
      
      gpc__queue_unref(die->qinfo);
      
      free(die->qname);
      g_free(die);

      lps_end(sys);
      
      return (g_realloc(qlist, sizeof(GPC_QUEUE_LIST *) * len));
    }
    
    ++scan;
  }

  g_assert_not_reached();
  
  lps_end(sys);
  
  return (qlist);
}

static void set_str(LPS_Printer_t *prn, const wchar_t *key, char *value){
  wchar_t *wtmp;

  if (value && *value)
  {
    wtmp=lps_promote(value);
    assert(wtmp);
    lps_pr_update_field(prn,key,wtmp);
    free(wtmp);
  }
}

static void set_bool(LPS_Printer_t *prn, const wchar_t *key, int test)
{
  if (test)
    lps_pr_update_field(prn, key, L"");
}

static void add_new_names(GPC_QUEUE_INFO *qinfo, LPS_Printer_t *prn)
{
  wchar_t **cur = NULL;
  char **curname = NULL;
  wchar_t *wtmp = NULL;
  
    // find the new names and add them to the list
  for(curname=qinfo->qlist;*curname!=NULL;curname++) {
    int matched=0;
    wtmp=lps_promote(*curname);
    for(cur=prn->names; *cur != NULL; cur++){
      if(!wcscmp(wtmp,*cur)){
	matched=1;
	break;
      }
    }

    if(matched==0){
      int vecsize=cur-prn->names;
      prn->names=realloc(prn->names,(vecsize+2) * sizeof(wchar_t **));
      prn->names[vecsize]=wtmp;
      prn->names[vecsize+1]=NULL;
    }else
      free(wtmp);
  }
}

static void sync_gpc2lps(GPC_QUEUE_INFO *qinfo, LPS_Printer_t *prn)
{
  GString *str = g_string_new("");

  assert(qinfo->spool_dir);
  set_str(prn,L"sd",qinfo->spool_dir);
  
  g_string_sprintf(str,"%d", qinfo->file_limit);
  set_str(prn,L"mx",str->str);
  set_bool(prn,L"if", qinfo->suppress_headers);
  set_str(prn,L"if",qinfo->ifilter);
  set_str(prn,L"af",qinfo->acct_file);
  set_str(prn,L"of",qinfo->ofilter);
  set_str(prn,L"lp",qinfo->local_device);
  set_str(prn,L"rm",qinfo->remote_host);
  set_str(prn,L"pr",qinfo->remote_queue);
  set_str(prn,L"printer_ip",qinfo->direct_host);
  if (qinfo->direct_port)
  {
    g_string_sprintf(str,"%d",qinfo->direct_port);
    set_str(prn,L"port",str->str);
  }
  set_str(prn,L"smb_hostip",qinfo->SMB_ip);
  set_str(prn,L"smb_host",qinfo->SMB_host);
  set_str(prn,L"smb_workgroup",qinfo->SMB_workgroup);
  if (qinfo->SMB_host && qinfo->SMB_printer_name)
  {
    g_string_sprintf(str,"//%s/%s",qinfo->SMB_host,qinfo->SMB_printer_name);
    set_str(prn,L"smb_share",str->str);
  }
  set_str(prn,L"user",qinfo->user);
  set_str(prn,L"passwd",qinfo->passwd);
  set_str(prn,L"ncp_server",qinfo->NCP_printserver_name);
  set_str(prn,L"ncp_queue",qinfo->NCP_queue_name);

  if (qinfo->filter_opts)
  {
    int ps_printer = FALSE;

    set_str(prn, L"PRINTER_MODEL", qinfo->filter_opts->printer_model);
    set_str(prn, L"GSDEVICE", qinfo->filter_opts->gs_dev);
    if (!strcmp(qinfo->filter_opts->gs_dev, "POSTSCRIPT"))
      ps_printer = TRUE;
    else
      ps_printer = FALSE;

    set_str(prn, L"RESOLUTION", qinfo->filter_opts->res);
    set_str(prn, L"PAPERSIZE", qinfo->filter_opts->size);
    set_str(prn, L"BITSPERPIXEL", qinfo->filter_opts->bpp);
    set_str(prn, L"RTLFTMAR", qinfo->filter_opts->margin_right_left);
    set_str(prn, L"TOPBOTMAR", qinfo->filter_opts->margin_top_bottom);
    set_str(prn, L"NUP", qinfo->filter_opts->pages_per_page);
    set_bool(prn, ps_printer ? L"TEXT_SEND_EOF" : L"PS_SEND_EOF",
             qinfo->filter_opts->send_eof);
    set_bool(prn, L"CRLFTRANS", qinfo->filter_opts->fix_stair_text);
    if (0)
    set_bool(prn, L"", qinfo->filter_opts->fast_text); /* ? */
    set_str(prn, L"EXTRA_GS_OPTIONS", qinfo->filter_opts->gs_opts);

  }
  
  g_string_free(str, TRUE);
}

GPC_QUEUE_LIST **gpc_queue_add(GPC_QUEUE_LIST **qlist, 
			       GPC_QUEUE_INFO *qinfo)
{
  LPS_System_t *sys=lps_init(printcap_filename);
  wchar_t **names=lps_promote_vec((const char**)qinfo->qlist);
  LPS_Printer_t *prn;
  size_t len = gpc__queue_len(qlist);
  
  qlist = g_realloc(qlist, (len + 2) * sizeof(GPC_QUEUE_LIST *));

  q_ret = qlist; /* FIXME: */
  
  qlist[len] = g_malloc(sizeof(GPC_QUEUE_LIST));
  qlist[len]->qname = g_strdup(qinfo->qlist[0]);
  qlist[len]->qinfo = qinfo;
  
  qlist[len + 1] = NULL;
  
  assert(sys);
  assert(names);
  prn=lps_create_printer(sys,names,NULL,NULL);

  sync_gpc2lps(qinfo, prn);
  
  lps_commit_printer(prn,NULL);

  lps_end(sys);

  gpc__queue_add(qlist, qinfo);
  
  return (qlist);
}

GPC_QUEUE_LIST **gpc_queue_edt(GPC_QUEUE_LIST **qlist, 
                               GPC_QUEUE_INFO *qinfo)
{
  LPS_System_t *sys = lps_init(printcap_filename);
  wchar_t *wtmp = lps_promote(qinfo->qlist[0]); /* FIXME: ? */
  LPS_Printer_t *prn = NULL;
  unsigned int scan = 0;
  int done = FALSE;
  
  assert(sys);
  assert(wtmp);
  prn=lps_get_printer(sys,wtmp,NULL);
  free(wtmp);
  assert(prn);
  if(!prn){
    wchar_t **names=lps_promote_vec((const char**)qinfo->qlist);
    assert(names);
    prn=lps_create_printer(sys,names,NULL,NULL);    
  }

  add_new_names(qinfo, prn);
  
  del_old_names(qinfo, prn);

  sync_gpc2lps(qinfo, prn);

  lps_commit_printer(prn,NULL);
  lps_end(sys);

  while (qlist[scan])
  {
    unsigned int inner_scan = 0;
    
    while (qinfo->qlist[inner_scan])
    {
      if (!strcmp(qlist[scan]->qname, qinfo->qlist[inner_scan]))
      {
        gpc__queue_unref(qlist[scan]->qinfo);
        qlist[scan]->qinfo = qinfo;
        gpc__queue_ref(qinfo);
        done = TRUE;
      }
      
      ++inner_scan;
    }  

    ++scan;
  }

  g_assert(done);
  
  return (qlist);
}

static int gpc__qsort_cmp(const void *t1, const void *t2)
{
  const char *const *const c1 = t1;
  const char *const *const c2 = t2;

  return (strcmp(*c1, *c2));
}

/* returns all names in a char *, Ie. join '|' */
char *gpc_queue_names(GPC_QUEUE_INFO *qinfo)
{
  GString *str = g_string_new("");
  unsigned int scan = 0;
  char *ret = NULL;
  size_t len = gpc__queue_node_len(qinfo);
  
  if (len)
  {
    char **tmp = g_malloc(sizeof(char *) * (len + 1));

    memcpy(tmp, qinfo->qlist, sizeof(char *) * (len + 1));
    
    qsort(tmp, len, sizeof(char *), gpc__qsort_cmp);
  
    assert(tmp[scan]);
    while (tmp[scan])
    {
      g_string_sprintfa(str, "%s%s", scan ? "|" : "", tmp[scan]);
      ++scan;
    }

    g_free(tmp);
  }
  
  ret = str->str;
  g_string_free(str, FALSE);

  return (ret);
}

GPC_QUEUE_LIST **gpc_queue_alias(GPC_QUEUE_LIST **qlist, 
				 GPC_QUEUE_INFO *qinfo,char *qname)
{
  LPS_System_t *sys = lps_init(printcap_filename);
  LPS_Printer_t *prn = NULL;
  wchar_t *wtmp = lps_promote(qinfo->qlist[0]);
  unsigned int scan = 0;
  size_t len = gpc__queue_len(q_ret);
  
  assert(sys);
  assert(wtmp);
  prn=lps_get_printer(sys,wtmp,NULL);
  free(wtmp);
  assert(prn);

  scan = gpc__queue_node_len(qinfo);

  qinfo->qlist = g_realloc(qinfo->qlist, sizeof(char *) * (scan + 2));
  qinfo->qlist[scan] = strdup(qname);
  qinfo->qlist[scan + 1] = NULL;

  add_new_names(qinfo, prn);

  lps_commit_printer(prn,NULL);
  lps_end(sys);

  qlist = gpc__queue_qlist_add_node(qlist, len, qname, qinfo);

  return (qlist);
}
