/* Copyright (c) 2000 Ben Woodard
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version 2
 * of the Licence, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRENTY; without even the implied warrenty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public Licence in the COPYING file for more
 * details.
 *
 * You should have received a copy of the GNU Library General 
 * Public License along with the GNU C Library; see the file 
 * COPYING.LIB.  If not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <syslog.h>
#include <tdb.h>
#include <unistd.h>

#include <glib.h>

#include "../config.h"
#include "../libprintsys/src/printsys.h"

#define MAX_PENDING 20
#define MAX_ATTRIB 16384
#define MAX_PRNAME 128
#define MAX_NS 128

#define CLIENT_TIMEOUT 2 // ititiation

#define QMGR_NAME "qmgr"
#define SENDLPR_NAME "sendlpr"

#define EXIT_SOCKET 1
#define EXIT_BIND 2
#define EXIT_LISTEN 12
#define EXIT_ACCEPT 4
#define EXIT_PASSCRED 5
#define EXIT_BADREAD 6
#define EXIT_BADPRINTER 7
#define EXIT_NODATA 8
#define EXIT_NODOCDB 9
#define EXIT_ATTRIBSIZE 10
#define EXIT_NOATTRIB 11
#define EXIT_MISCONFIG 13
#define EXIT_SVRERR 14
#define EXIT_CREDS 15
#define EXIT_TOREAD_ERR 16
#define EXIT_ATTRIBLEN 17
#define EXIT_ATTRSYNTAX 18
#define EXIT_BADUSER 19
#define EXIT_NOERRFD 20
#define EXIT_WAIT 21
#define EXIT_FILTERFAIL 22
#define EXIT_PIPELINE 23
#define EXIT_FORK1 24
#define EXIT_FORK2 25
#define EXIT_SIGNAL 26
#define EXIT_MALLOC 27
#define EXIT_TDBERR 28
#define EXIT_NOAVAILJOBS 29

typedef enum {CMD_ERR_EOF,CMD_ERR_SELECT,CMD_ERR_SELUNK,CMD_ERR_READ,
	      CMD_ERR_BADCMD,CMD_ERR_ANCSHORT,CMD_ERR_BADLVL,CMD_ERR_BADTYP,
	      CMD_ERR_NSLONG,
	      CMD_SUB,CMD_LST,CMD_PRN} Cmd_t;

/* This structure is used by get_unique_jobid and is inserted in 
   the tdb database to indicate a block of jobid's that is 
   completely filled. All the integers are signifigant and so to 
   make this key unique we make it a little bit bigger than an 
   integer by padding the structure with a single char. That 
   single char is always set to 0 */
struct block_field{
  char blank; // this is always 0. It is just there for padding
  int block;
};

static void daemon_init(const char *pname, int facility);
static void take_connection(int fd,int errfd,struct ucred *creds,
			    TDB_CONTEXT *tdb,LPS_Printer_t *printer);
static void take_printer_list_request(int fd,struct ucred *creds, 
				      LPS_System_t *sys,char *ns);
static void take_printer_request(int fd,struct ucred *creds,
				 LPS_Printer_t *printer);
static unsigned int get_unique_jobid(TDB_CONTEXT *tdb,
				     LPS_Pair_t *attributes);
static gint alpha_sort(gconstpointer a, gconstpointer b);
static void master_sig_chld_handler(int i);
static ssize_t timed_read(int fd, void *buf, size_t count,time_t timeout);
static LPS_Pair_t *read_attributes(int fd,struct ucred *creds);
static LPS_Pair_t *purify_attributes(LPS_Pair_t *attributes,
				     struct ucred *creds,int fd);
static GSList *get_filter_list(LPS_Printer_t *printer, int *endpoint, 
			       int fd);
static Cmd_t read_cmd(int fd,wchar_t *printer,ssize_t prlen,time_t timeout,
		      struct ucred *creds,char *ns,ssize_t nslen, 
		      int *errfd);
static void event_loop(LPS_System_t *sys, TDB_CONTEXT *tdb,int sockfd,
		       socklen_t client_len, struct sockaddr_un* addr);

/********** Global Variables **********/
char master=1;

/********** Main **********/
int main(int argc,char **argv){
  const char defsocket[]=LPS_SOCK_PATHNAME;
  const char defdocdb[]=LPS_DOCDBPATH;
  int sockfd;
  struct sockaddr_un *addr;
  socklen_t addr_len = 0;
  TDB_CONTEXT *tdb;
  static const struct option options[]= { 
    {"version",no_argument,      NULL,'V'},
    {"help",   no_argument,      NULL,'?'},
    {"debug",  no_argument,      NULL,'d'}, // primarily for debugging
    {"socket", required_argument,NULL,'s'}, // primarily for debugging
    {"jobdb",  required_argument,NULL,'j'}, // primarily for debugging
    {NULL,0,NULL,0}
  };
  int optval;
  char daemonize=1;
  const char *sockname=defsocket;
  size_t sockname_len = 0;
  const char *docdb=defdocdb;
  LPS_System_t *sys;

  while((optval=getopt_long(argc,argv,"VHdps:j:",options,NULL))!=EOF){
    switch(optval){
    case 'V':
      printf("lpd Version: %s",VERSION);
      exit(0);
    case 'd':
      daemonize=0;
      break;
    case 's':
      sockname=optarg;
      break;
    case 'j':
      docdb=optarg;
      break;
    case '?':
      printf("usage: lpd -V or\nlpd [-d]\n");
      exit(0);
    }
  }

  if(daemonize)
    daemon_init("lpd",LOG_LPR);

  syslog(LOG_INFO,"Starting lpd");

  if((sys=lps_init(NULL))==NULL){
    syslog(LOG_CRIT,"Cannot initialize print system");
    exit(EXIT_NODATA);
  }

  syslog(LOG_DEBUG,"%s",docdb);
  if((tdb=tdb_open(docdb,0,0,O_RDWR|O_CREAT,0644))==NULL){
    syslog(LOG_CRIT,"Cannot open jobid database %s: %s",docdb,
	   sys_errlist[errno]);
    exit(EXIT_NODOCDB);
  }

  syslog(LOG_DEBUG,"tdb.header.version=%d",tdb->header.version);
  tdb->header.version=637606253;

  if(signal(SIGCHLD,master_sig_chld_handler)==SIG_ERR){
    syslog(LOG_CRIT,"Cannot install signal handler");
    exit(EXIT_SIGNAL);
  }

  if((sockfd=socket(AF_LOCAL,SOCK_STREAM,0))==-1){
    syslog(LOG_CRIT,"Cannot open a socket: %s",sys_errlist[errno]);
    exit(EXIT_SOCKET);
  }
  
  unlink(sockname); // ok to fail

  sockname_len = strlen(sockname);
  addr_len = sizeof(struct sockaddr_un) + sockname_len + 1;
  if (!(addr = malloc(addr_len))){
    syslog(LOG_CRIT,"Cannot create space to store a socket");
    exit(EXIT_SOCKET);
  }
  
  /* Only need to blank the struct part */
  memset(addr, 0, sizeof(struct sockaddr_un));
  addr->sun_family = AF_LOCAL;
  memcpy(addr->sun_path, sockname, sockname_len + 1);
  
  if(bind(sockfd,(struct sockaddr*)addr, SUN_LEN(addr))==-1){
    syslog(LOG_CRIT,"Cannot bind socket %s: %s", sockname,sys_errlist[errno]);
    exit(EXIT_BIND);
  }

  if(listen(sockfd,MAX_PENDING)==-1){
    syslog(LOG_CRIT,"Cannot listen on socket: %s", sys_errlist[errno]);
    exit(EXIT_LISTEN);
  }
  event_loop(sys,tdb,sockfd,addr_len,addr);
  return 0;
}

void event_loop(LPS_System_t *sys, TDB_CONTEXT *tdb,int sockfd,
		socklen_t client_len, struct sockaddr_un* addr){
  for(;;){
    int connfd,errfd;
    pid_t pid;
    wchar_t buf[MAX_PRNAME+1];
    char ns[MAX_NS+1];
    LPS_Printer_t *printer;
    struct ucred creds;
    int val=1;
    Cmd_t cmd;

    if((connfd = accept(sockfd,(struct sockaddr*)addr,&client_len))==-1){
      syslog(LOG_ERR,"Cannot accept a pending connection: %s", 
	     sys_errlist[errno]);
      continue;
    }

    if (addr->sun_family != AF_LOCAL){
     /* This is also possible via. connect()/close() */
      syslog(LOG_ERR,"Non valid fd from accept");
      close(connfd);
      continue;
    }
    
    if (setsockopt(connfd,SOL_SOCKET, SO_PASSCRED, &val, sizeof(int)) == -1){
      syslog(LOG_ERR,"Cannot set socket option PASSCRED: %s", 
	     sys_errlist[errno]);
      close(connfd);
      continue;
    }
    /* Finding out the name of the printer here is sort of an 
       optimization with the way that I implemented the printcap 
       parser and how I am likely to implement the other parsers.
       This way the lp_init reads the file once and then every 
       lookup does a stat on the file to see if is up to date. 
       That way there is as little reparsing of a printcap file 
       as possible. 

       If I did the lp_init in the forked child every printjob 
       would have to read the printcap file.

       If I did the lp_init in the main process and then did the 
       lookup in the forked child if the printcap ever changed 
       then every child would have to reparse it. */
    buf[MAX_PRNAME]=0;
    cmd=read_cmd(connfd,buf,MAX_PRNAME,CLIENT_TIMEOUT,&creds,ns,MAX_NS,
		 &errfd);
    if(cmd!=CMD_SUB && cmd!=CMD_LST && cmd!=CMD_PRN ){
      static const char *errstr;
      switch(cmd){
      case CMD_ERR_EOF:
	errstr="premature end of file";
	break;
      case CMD_ERR_SELECT:
	errstr="select returned error";
	break;
      case CMD_ERR_SELUNK:
	errstr="select returned unknown error";
	break;
      case CMD_ERR_READ:
	errstr="read returned an error";
	break;
      case CMD_ERR_BADCMD:
	errstr="unknown command";
	break;
      case CMD_ERR_ANCSHORT:
	errstr="not enough ancillary data";
	break;
      case CMD_ERR_BADLVL:
	errstr="ancillary data not on right level";
	break;
      case CMD_ERR_BADTYP:
	errstr="ancillary data not the right type";
	break;
      case CMD_ERR_NSLONG:
	errstr="ns string too long";
	break;
      default:
	errstr="unknown error";
	break;
      }
      close(connfd);
      syslog(LOG_DEBUG,"bad command: %s",errstr);
      continue;
    }

    if(cmd==CMD_SUB || cmd==CMD_LST){
      if((printer=lps_get_printer(sys,buf+4,NULL))==NULL){
	const char *tmp="Err 1: Bad Printer Name\n";
	write(connfd,tmp,strlen(tmp));
	syslog(LOG_NOTICE,"Client with id:%d requested invalid printer with "
	       "name: \"%s\"\n",creds.uid,buf);
	close(connfd);
	continue;
      }
    }
    
    if((pid=fork())==-1){
      const char *tmp="IntErr 0: Cannot fork to handle job\n";
      syslog(LOG_CRIT,"Cannot fork to handle pending connection: %s", 
	     sys_errlist[errno]);
      write(connfd,tmp,strlen(tmp));
      close(connfd);
      continue;
    }

    if(pid!=0){ // master server done with its work
      close(connfd);
      continue;
    }

    master=0;
    close(sockfd);
    syslog(LOG_DEBUG, "taking connection");

    if(cmd==CMD_SUB){
      write(connfd,"OK\n",3);
      take_connection(connfd,errfd,&creds,tdb,printer);
    } else if(cmd==CMD_LST)
      take_printer_list_request(connfd,&creds,sys,ns);
    else if(cmd=CMD_PRN)
      take_printer_request(connfd,&creds,printer);
    exit(EXIT_OK);
  } // infinite loop
}

/* Hat tip to W. Richard Stevens Stolen right out of UNP V1 2ed */
static void daemon_init(const char *pname, int facility){
  int i;
  pid_t	pid;

  if((pid=fork())==-1){
    /* This is a fprintf because we won't open the log until we 
       get to the end of this function. */
    fprintf(stderr,"daemon_init: Fork 1 failed: %s", sys_errlist[errno]);
    exit(EXIT_FORK1);
  }
  if(pid!=0)
    exit(0); // parent terminates

  /* 1st child continues */
  setsid(); // become session leader

  signal(SIGHUP, SIG_IGN);
  if((pid=fork())==-1){
    fprintf(stderr,"daemon_init: Fork 2 failed: %s", sys_errlist[errno]);
    exit(EXIT_FORK2);
  }
  if(pid!=0)
    exit(0); // 1st child terminates

  /* 2nd child continues */
  chdir("/"); // change working directory

  umask(0); //clear our file mode creation mask

  for(i=0;i<3;i++)
    close(i);

  openlog(pname, LOG_PID, facility);
}

static gint alpha_sort(gconstpointer a, gconstpointer b){
  return strcmp((char*)a,(char *)b);
}

static unsigned int get_unique_jobid(TDB_CONTEXT *tdb,
				     LPS_Pair_t *attrib){
  struct block_field block;
  TDB_DATA block_key;
  unsigned int i;
  size_t buflen;
  char *buf=_lps_encode_pairs(attrib,&buflen);

  syslog(LOG_DEBUG,"get_unique_jobid attrib=%p",attrib);

  block.blank=0; // this will always be 0
  // skip over filled blocks
  block_key.dptr=(char*)&block;
  block_key.dsize=sizeof(block);
  for(block.block=0;tdb_exists(tdb,block_key);block.block+=100)
    syslog(LOG_DEBUG,"block %d full",block.block);
  // find a unique key.
  for(i=block.block;;i++){
    TDB_DATA attributes;
    TDB_DATA job_key;
    int tdbret;

    syslog(LOG_DEBUG,"trying %u",i);
    attributes.dptr=buf;
    attributes.dsize=buflen; 
    job_key.dptr=(char*)&i;
    job_key.dsize=sizeof(i);
    syslog(LOG_DEBUG,"key dsize=%d dptr[0]=%02x dptr[1]=%02x dptr[2]=%02x "
	   "dptr[3]=%02x",job_key.dsize,job_key.dptr[0],job_key.dptr[1],
	   job_key.dptr[2],job_key.dptr[3]);
    
    if((tdbret=tdb_store(tdb,job_key,attributes,TDB_INSERT))!=-1){
      syslog(LOG_DEBUG,"tdbret=%d",tdbret); 
      syslog(LOG_DEBUG,"returning=%u",i); 
      return i;
    }
    syslog(LOG_DEBUG,"tdbret=%d",tdbret); 
    if(tdb_error(tdb)!=TDB_ERR_EXISTS){
      syslog(LOG_ALERT,"Can't store in tdb \"%s\": %s",tdb_errorstr(tdb),
	     sys_errlist[errno]);
      exit(EXIT_TDBERR);
    }
    // if the entire block is used mark it as full
    if((i+1)%100==0){ 
      tdb_store(tdb,block_key,attributes,TDB_REPLACE);
      block.block+=100;
    }
  }
  syslog(LOG_ALERT,"Out of print job slots. Exiting.");
  exit(EXIT_NOAVAILJOBS);
}

static void master_sig_chld_handler(int i __attribute__ ((unused))){
  pid_t child_pid;
  int status;
  for(;;){
    if((child_pid=waitpid(-1,&status,WNOHANG))==0) // no zombies left
      return;
    if(child_pid==-1){
      if(master){
	if(errno==ECHILD)
	  return;
	syslog(LOG_ALERT,"Error waiting for a child: %s\nMaster server "
	       "terminating\n",sys_errlist[errno]);
	exit(EXIT_WAIT);
      }else{
	syslog(LOG_ERR,"Error waiting for filter: %s\nDaemon terminating\n",
	       sys_errlist[errno]);
	return;
	// exit(EXIT_WAIT);
      }
    }

    if(WIFEXITED(status)){
      int exitval=WEXITSTATUS(status);
      if(exitval==0){
	continue;
      } else {
	if(master){
	  syslog(LOG_WARNING,"Child daemon exited with a status of %d", 
		 exitval);
	  continue;
	}else{
	  syslog(LOG_ERR,"Filter exited with a status of %d", exitval);
	  exit(EXIT_FILTERFAIL);
	}
      }
    }

    if(WIFSIGNALED(status)){
      if(master){
	syslog(LOG_ERR,"Child daemon pid %d was killed by an uncaught "
	       "signal %d", child_pid,WTERMSIG(status));
	continue;
      }else{
	syslog(LOG_ERR,"Filter process %d was killed by an uncaught signal"
	       " %d", child_pid,WTERMSIG(status));
	exit(EXIT_FILTERFAIL);
      }
    }
  } // end of for
} // end of signal handler

static ssize_t timed_read(int fd, void *buf, size_t count, time_t timeout){
  fd_set ifds,ofds;
  struct timeval tv;
  int sel_fd;
  FD_ZERO(&ifds);
  FD_ZERO(&ofds);
  FD_SET(fd,&ifds);
  tv.tv_sec=timeout;
  tv.tv_usec=0;
  
  switch(sel_fd=select(fd+1,&ifds,&ofds,&ofds,&tv)){
  case 0:
    errno=EAGAIN;
    // intentionally falls through
  case -1:
    return -1;
  case 1:
    return read(fd,buf,count);
  }

  // This is essentially the default case.
  syslog(LOG_ERR,"timed_read: select returned an invalid value, %d, exiting",
	 sel_fd);
  exit(EXIT_TOREAD_ERR);
}

static LPS_Pair_t *read_attributes(int fd,struct ucred *creds){
  ssize_t attrib_size;
  ssize_t len_read=timed_read(fd,&attrib_size,sizeof(attrib_size),
			     CLIENT_TIMEOUT);
  char *buf;
  LPS_Pair_t *attributes=NULL;
  char *curbuf;
  int bytesleft;

  if(len_read==-1){
    syslog(LOG_WARNING,"Could not read attribute size: %s. Connection "
	   "dropped. Client uid=%d",sys_errlist[errno],creds->uid);
    exit(EXIT_ATTRIBSIZE);
  }else if(len_read==0){
    syslog(LOG_WARNING,"Connection to client closed before sending "
	   "attributes. Client uid=%d",creds->uid);
    exit(EXIT_NOATTRIB);
  }else if(len_read!=sizeof(attrib_size)){
    static const char str[]="ERR 1: Attribute size wrong length\n";
    syslog(LOG_WARNING,"Client didn't send enough bytes for attrib_size. "
	   "Client uid=%d", creds->uid);
    write(fd,str,strlen(str));
    exit(EXIT_ATTRIBSIZE);
  }

  if(attrib_size==0)
    return attributes;

  if(attrib_size>MAX_ATTRIB){
    static const char str[]="Err 2: Attribute size is too big\n";
    syslog(LOG_WARNING,"Client tried to request too big of an attribute "
	   "size. Client uid=%d", creds->uid);
    write(fd,str,strlen(str));
    exit(EXIT_ATTRIBSIZE);
  }
  if((buf=malloc(attrib_size))==0){
    static const char str[]="Err 3: Buffer space for attributes could not "
      "be allocated.\n";
    syslog(LOG_WARNING,"Attribute buffer of size %d size could not be "
	   "allocated. Client uid=%d", attrib_size, creds->uid);
    write(fd,str,strlen(str));
    exit(EXIT_NOMEM);
  }

  // read in the attributes
  curbuf=buf;
  bytesleft=attrib_size;
  do{
    len_read=timed_read(fd,curbuf,bytesleft,CLIENT_TIMEOUT); 
    // need to have some error handling code in here to deal with 
    // timouts and failed reads etc.
    curbuf-=len_read;
    bytesleft-=len_read;
  }while(bytesleft>0);

  syslog(LOG_DEBUG,"converting attrib");
  if((attributes=_lps_decode_pairs(buf,attrib_size,attributes))==NULL){
    static const char str[]="Err 5: Syntax error in the structure of "
      "attributes.";
    syslog(LOG_WARNING,"Syntax error in the structure of attributes. Client "
	   "uid=%d",creds->uid);
    write(fd,str,strlen(str));
    exit(EXIT_ATTRSYNTAX);
  }
  free(buf);
  return attributes;
}

static LPS_Pair_t *purify_attributes(LPS_Pair_t *attributes,
				     struct ucred *creds,int fd){
  GString *tmpstr=g_string_new("");
  wchar_t *tmpwcs;
  wchar_t *key;

  // insert the user's credentials overwriting any that the user
  // may have tried to pass in illicitly.
  if(lps_pair_lookup(attributes,LPS_DOC_UID)!=NULL){
    static const char str[]="Err 6: Attempt to set UID";
    syslog(LOG_WARNING,"Client uid=%d attempted to set UID", creds->uid);
    write(fd,str,strlen(str));
    exit(EXIT_BADUSER);
  }
  if(lps_pair_lookup(attributes,LPS_DOC_GID)!=NULL){
    static const char str[]="Err 7: Attempt to set GID";
    syslog(LOG_WARNING,"Client uid=%d attempted to set GID", creds->uid);
    write(fd,str,strlen(str));
    exit(EXIT_BADUSER);
  }
  if(lps_pair_lookup(attributes,LPS_DOC_PID)!=NULL){
    static const char str[]="Err 8: Attempt to set PID";
    syslog(LOG_WARNING,"Client uid=%d attempted to set PID", creds->uid);
    write(fd,str,strlen(str));
    exit(EXIT_BADUSER);
  }
  
  // insert the uid
  g_string_sprintf(tmpstr,"%d",creds->uid);
  tmpwcs=lps_promote(tmpstr->str);
  key=malloc((wcslen(LPS_DOC_UID)+1)*sizeof(wchar_t));
  wcscpy(key,LPS_DOC_UID);
  attributes=lps_pair_update(attributes,key,tmpwcs);

  // insert the gid
  g_string_sprintf(tmpstr,"%d",creds->gid);
  tmpwcs=lps_promote(tmpstr->str);
  key=malloc((wcslen(LPS_DOC_GID)+1)*sizeof(wchar_t));
  wcscpy(key,LPS_DOC_GID);
  attributes=lps_pair_update(attributes,key,tmpwcs);

  // insert the pid
  g_string_sprintf(tmpstr,"%d",creds->pid);
  tmpwcs=lps_promote(tmpstr->str);
  key=malloc((wcslen(LPS_DOC_PID)+1)*sizeof(wchar_t));
  wcscpy(key,LPS_DOC_PID);
  attributes=lps_pair_update(attributes,key,tmpwcs);
  g_string_free(tmpstr,TRUE);

  return attributes;
}

static GSList *insert_filter(GSList *filters,wchar_t *newone){
      char *curstr=lps_demote(newone);
      GString *str=g_string_new(curstr);
      filters=g_slist_append(filters,str->str);
      g_string_free(str,FALSE);
      free(curstr);
      return filters;
}

static GSList *get_filter_list(LPS_Printer_t *printer, int *endpoint,
			       int fd){
  static const char err4[]="IntErr 1: Internal error. Printer is "
    "misconfigured on the server.\n";
  wchar_t *cur;
  GSList *filters=NULL;

  syslog(LOG_DEBUG,"get_filter_list");

  if((cur=lps_pr_lookup_field(printer,L"fd"))){
    char *tmpstr=lps_demote(cur);
    DIR *filterdir=opendir(tmpstr);
    struct dirent *entry;

    if(filterdir==NULL){
      syslog(LOG_ERR,"Error opening filter directory for the printer"
	     "\"%ls\": %s",printer->names[0],sys_errlist[errno]);
      write(fd,err4,strlen(err4));
      exit(EXIT_MISCONFIG);
    }

    while((entry=readdir(filterdir))){
      GString *str;
      if(!strcmp(entry->d_name,".") || !strcmp(entry->d_name,".."))
	continue;
      str=g_string_new(tmpstr);
      g_string_sprintfa(str,"/%s",entry->d_name);
      // syslog(LOG_DEBUG,"Adding filter=(%s)",str->str);
      filters=g_slist_insert_sorted(filters,str->str,alpha_sort);
      g_string_free(str,FALSE);
    }
    free(tmpstr);
  }else{ // backward BSD compatability mode
    if((cur=lps_pr_lookup_field(printer,L"if"))!=NULL)
      filters=insert_filter(filters,cur);
    if((cur=lps_pr_lookup_field(printer,L"rm"))!=NULL){
      GString *str=g_string_new(LIBEXEC_PATH);
      g_string_sprintfa(str,"/%s",SENDLPR_NAME);
      filters=g_slist_append(filters,str->str);
      g_string_free(str,FALSE);
    }else {
      /* You don't need a queue manager when you are just 
	 forwarding it to another print server and you can't have 
	 an output filter with a remote machine. */
      GString *str=g_string_new(LIBEXEC_PATH);
      g_string_sprintfa(str,"/%s",QMGR_NAME);
      filters=g_slist_append(filters,str->str);
      g_string_free(str,FALSE);
      if((cur=lps_pr_lookup_field(printer,L"of"))!=NULL)
	filters=insert_filter(filters,cur);
      // Might need an endpoint but not with RM
      if((cur=lps_pr_lookup_field(printer,L"lp"))!=NULL){
	char *tmpstr=lps_demote(cur);
	*endpoint=open(tmpstr,O_RDWR);
	free(tmpstr);
      }
      if(*endpoint==-1){
	syslog(LOG_ERR,"Error opening printer device for printer \"%ls\": %s",
	       printer->names[0],sys_errlist[errno]);
	write(fd,err4,strlen(err4));
	exit(EXIT_MISCONFIG);
      }
    }
  } // end of BSD compatability mode


  if(*endpoint==-1){
    if((*endpoint=open("/dev/null",O_RDWR))==-1){
      static const char str[]="IntErr 2: Internal spooler error\n";
      syslog(LOG_ALERT,"Can't open null endpoint: %s",sys_errlist[errno]);
      write(fd,str,strlen(str));
      exit(EXIT_SVRERR);
    }
  }

  syslog(LOG_DEBUG,"finished get_filter_list");
  return filters;
}

static void take_connection(int fd,int errfd,struct ucred *creds,
			    TDB_CONTEXT *tdb,LPS_Printer_t *printer){
  LPS_Pair_t *attributes=purify_attributes(read_attributes(fd,creds),creds,fd);
  int endpoint=-1;
  GSList *filters=get_filter_list(printer,&endpoint,fd);
  GSList *curfilter;
  unsigned int jobid=get_unique_jobid(tdb,attributes);
  GString *jobid_str=g_string_new("");
  fd_set ifds,ofds;
  
  syslog(LOG_DEBUG,"jobid=%u",jobid); 
  g_string_sprintf(jobid_str,"%u",jobid);
  write(fd,jobid_str->str,jobid_str->len);
  write(fd,"\n",1);
  syslog(LOG_DEBUG,"data mode");
  FD_ZERO(&ifds);
  FD_ZERO(&ofds);
  FD_SET(fd,&ifds);
  /* From this point on everthing is expected to be print data
     and we can no-longer write to the fd because it might be 
     connected to a pipe. */
  
  if(select(fd+1,&ifds,&ofds,&ofds,NULL)!=1){
    GString *str=g_string_new("Failure when selecting on datastream:");
    syslog(LOG_WARNING,"Select before pipeline failed: %s",sys_errlist[errno]);
    g_string_sprintfa(str,"str, %s\n",sys_errlist[errno]);
    write(errfd,str->str,str->len);
    g_string_free(str,TRUE);
    exit(EXIT_PIPELINE);
  }

  syslog(LOG_DEBUG,"Ready to go");
  for(curfilter=filters;curfilter!=NULL;curfilter=g_slist_next(curfilter)){
    int pends[2];
    char *prname=lps_demote(printer->names[0]);
    syslog(LOG_DEBUG,"execing filter \"%s\" args=\"%s\" \"%s\" \"%s\"",
	   curfilter->data,curfilter->data,prname,jobid_str->str);
    syslog(LOG_DEBUG,"tdb.header.version=%d",tdb->header.version);
    if(g_slist_next(curfilter)==NULL)
      pends[1]=endpoint;
    else if(pipe(pends)!=0){
      GString *str=g_string_new("Failure when trying to create a new pipe:");
      syslog(LOG_WARNING,"Pipe creation failed: %s",sys_errlist[errno]);
      g_string_sprintfa(str," %s\n",sys_errlist[errno]);
      write(errfd,str->str,str->len);
      g_string_free(str,TRUE);
      exit(EXIT_PIPELINE);
    }
    switch(fork()){
    case -1:{
      GString *str=g_string_new("Fork failed when trying to construct "
				"pipeline:");
      syslog(LOG_WARNING,"Fork failed: %s",sys_errlist[errno]);
      g_string_sprintfa(str," %s\n",sys_errlist[errno]);
      write(errfd,str->str,str->len);
      g_string_free(str,TRUE);
      exit(EXIT_PIPELINE);
    }
    case 0:{
      GString *errstr=g_string_new("Printer configuration error. Could not "
				   "exec filter ");
      if( dup2(fd,0)==-1 || dup2(pends[1],1)==-1 || dup2(errfd,2)==-1){
	GString *str=g_string_new("Error while rearranging file descriptors:");
	syslog(LOG_WARNING,"Dup2 failed while trying to construct pipeline: "
	       "%s",sys_errlist[errno]);
	g_string_sprintfa(str," %s\n",sys_errlist[errno]);
	write(errfd,str->str,str->len);
	exit(EXIT_PIPELINE);
      }
      // close fd's if they happen to not be the ones we are using
      if(fd>2) 
	close(fd);
      if(pends[0]>2)
	close(pends[0]);
      if(pends[1]>2)
	close(pends[1]);
      if(errfd>2)
	close(errfd);
      // XX - might want to close more fds
      syslog(LOG_DEBUG,"doing exec");
      execl(curfilter->data,curfilter->data,prname,jobid_str->str,NULL);
      syslog(LOG_WARNING,"exec of filter %s failed %s",curfilter->data,
	     sys_errlist[errno]);
      g_string_sprintfa(errstr,"%s: %s\n", (char*) curfilter->data,
			sys_errlist[errno]);
      write(2,errstr->str,strlen(errstr->str));
      exit(EXIT_MISCONFIG);
    }
    default:
      close(pends[1]);
      dup2(pends[0],fd);
      close(pends[0]);
      if(g_slist_next(curfilter)==NULL)
	break;
      FD_ZERO(&ifds);
      FD_ZERO(&ofds);
      FD_SET(fd,&ifds);
      if(select(fd+1,&ifds,&ofds,&ofds,NULL)!=1){
	GString *str=g_string_new("Failure when selecting on datastream:");
	syslog(LOG_WARNING,"Select before pipeline failed: %s",
	       sys_errlist[errno]);
	g_string_sprintfa(str," %s\n",sys_errlist[errno]);
	write(errfd,str->str,str->len);
	g_string_free(str,TRUE);
	exit(EXIT_PIPELINE);
      }
      break;
    }
  }
  exit(EXIT_OK);
}

static Cmd_t read_cmd(int fd,wchar_t *printer,ssize_t prlen,time_t timeout,
		      struct ucred *creds,char *ns,ssize_t nslen, 
		      int *errfd){
  struct msghdr msg;
  struct iovec iov[1];
  union {
    struct cmsghdr cm; // seems to be only for alignment
    char control[CMSG_SPACE(sizeof (int))+CMSG_SPACE(sizeof(struct ucred))];
  } control_un;
  struct cmsghdr  *fdhdr,*credhdr;
  fd_set ifds,ofds;
  struct timeval tv;
  int err_fd;
  ssize_t tot_bytes;
  int save_errno;
  ssize_t actual_nslen;
  
  memset(printer,0,prlen);

  FD_ZERO(&ifds);
  FD_ZERO(&ofds);
  FD_SET(fd,&ifds);
  tv.tv_sec=timeout;
  tv.tv_usec=0;
 
  iov[0].iov_base = printer;
  iov[0].iov_len = prlen*sizeof(wchar_t);

  msg.msg_control = control_un.control;
  msg.msg_controllen = sizeof(control_un.control);
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = iov;
  msg.msg_iovlen = 1;

  switch(err_fd=select(fd+1,&ifds,&ofds,&ofds,&tv)){
  case 0:
    syslog(LOG_WARNING,"read_printer: client didn't send printer name "
	   "expiediently. Dropping connection.");
    return CMD_ERR_EOF;
  case -1:
    save_errno=errno;
    syslog(LOG_WARNING,"read_printer: error waiting for printer name: %s",
	   sys_errlist[errno]);
    errno=save_errno;
    return CMD_ERR_SELECT;
  case 1:
    break;
  default:
    syslog(LOG_WARNING,"read_printer: select returned an invalid value, %d. "
	   "Exiting", err_fd);
    return CMD_ERR_SELUNK;
  }

  switch(tot_bytes=recvmsg(fd,&msg,0)){
  case 0:
    return CMD_ERR_EOF;
  case -1:
    save_errno=errno;
    syslog(LOG_WARNING,"read_printer: could not read printer: %s",
	   sys_errlist[errno]);
    errno=save_errno;
    return CMD_ERR_READ;
  }

  /* XXX: This is a hack to work around a bug in glibc where 
     CMSG_NXTHDR returns NULL even though it has enough data */
  msg.msg_controllen++;

  credhdr=CMSG_FIRSTHDR(&msg); 
  fdhdr=CMSG_NXTHDR(&msg,credhdr);

  // swap them around if necessary
  if(fdhdr && fdhdr->cmsg_type==SCM_CREDENTIALS && 
     credhdr->cmsg_type==SCM_RIGHTS){
    fdhdr=CMSG_FIRSTHDR(&msg);
    credhdr=CMSG_NXTHDR(&msg,fdhdr);
  }

  if(credhdr==NULL){
    syslog(LOG_WARNING,
	   "read_printer: not enough ancillary data for creds -- exiting.");
    return CMD_ERR_ANCSHORT;
  }

  if(credhdr->cmsg_level!=SOL_SOCKET ){
    syslog(LOG_WARNING,"read_printer: cred ancillary data not on the correct "
	   "level  -- exiting.");
    return CMD_ERR_BADLVL;
  }

  if(credhdr->cmsg_type!=SCM_CREDENTIALS){
    syslog(LOG_WARNING,"read_printer: ancillary data not of the right type. "
	   "-- exiting");
    return CMD_ERR_BADTYP;
  }

  memcpy(creds,CMSG_DATA(credhdr),sizeof(*creds));

  if(!strncmp(iov[0].iov_base,"sub\n",4)){
    int *foo;
    if(fdhdr==NULL){
      syslog(LOG_WARNING,
	     "read_printer: not enough ancillary data for errror -- exiting.");
      return CMD_ERR_ANCSHORT;
    }

    if(fdhdr->cmsg_level!=SOL_SOCKET){
      syslog(LOG_WARNING,"read_printer: fd ancillary data not on the correct "
	     "level  -- exiting.");
      return CMD_ERR_BADLVL;
    }

    if(fdhdr->cmsg_type!=SCM_RIGHTS){
      syslog(LOG_WARNING,"read_printer: fd ancillary data not of the right "
	     "type. -- exiting");
      return CMD_ERR_BADTYP;
    }
  
    foo=(int*)CMSG_DATA(fdhdr);
    err_fd=*foo;
    memcpy(iov[0].iov_base,iov[0].iov_base+4,tot_bytes-4);
    return CMD_SUB;
  }

  if(strncmp(iov[0].iov_base,"lst\n",4) && 
       strncmp(iov[0].iov_base,"prn\n",4)){
    return CMD_ERR_BADCMD;
  }

  actual_nslen=(ssize_t) *(printer+4);
  if(actual_nslen+1>nslen)
    return CMD_ERR_NSLONG;
  memset(ns,0,nslen);
  memcpy(ns,iov[0].iov_base+4+sizeof(ssize_t),actual_nslen);

  if(!strncmp(iov[0].iov_base,"lst\n",4))
    return CMD_LST;

  memcpy(iov[0].iov_base, iov[0].iov_base+4+sizeof(ssize_t)+actual_nslen,
	 tot_bytes-4-sizeof(ssize_t)-actual_nslen);
  return CMD_PRN;
}

static void take_printer_list_request(int fd,struct ucred *creds 
				      __attribute((unused)), 
				      LPS_System_t *sys,char *ns){
  const wchar_t **prlist=lps_get_printer_list(sys,ns);
  const wchar_t **cur;
  GString *str=g_string_new("");

  // could log the credentials of the person who made the request 
  // here.

  for(cur=prlist;*cur!=NULL;cur++)
    g_string_sprintfa(str,"%S\n",*cur);
  g_string_sprintfa(str,"\000");
  write(fd,str->str,str->len);
}

static void take_printer_request(int fd,struct ucred *creds,
				 LPS_Printer_t *printer){
  char *str=_lps_write_printcap_entry(printer);
  write(fd,str,strlen(str));
  free(str);
}


