/* Copyright (c) 2000 Ben Woodard
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version 2
 * of the Licence, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRENTY; without even the implied warrenty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public Licence in the COPYING file for more
 * details.
 *
 * You should have received a copy of the GNU Library General 
 * Public License along with the GNU C Library; see the file 
 * COPYING.LIB.  If not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <unistd.h>
#include <wchar.h>
#include "../libprintsys/src/printsys.h"

#include <glib.h>

#define EXIT_SOCK 4
#define EXIT_BADRM 5
#define EXIT_NOCONNECT 6
#define EXIT_RELAYSOCK 7
#define EXIT_BINDRELAY 8
#define EXIT_LISTENRELAY 9
#define EXIT_RELAYSOCK2 10
#define EXIT_NOPREV 11
#define EXIT_NOHANDOFF 12
#define EXIT_BADRELAY 13
#define EXIT_RELAYDUP 14
#define EXIT_BADCONN 15
#define EXIT_BIGJOB 16
#define EXIT_DATAFILE 17
#define EXIT_DFCOMMAND 18
#define EXIT_RELAYACCEPT 19
#define EXIT_RELAYWRITE 20
#define EXIT_NORM 21
#define EXIT_NORP 22
#define EXIT_NOSD 23
#define EXIT_BADSD 24
#define EXIT_JOBSIZE 25
#define EXIT_DOCFILE 26
#define EXIT_SPOOLFILE 27
#define EXIT_DUPDOCFILE 28
#define EXIT_PREVRELAY 29
#define EXIT_HANDOFF 30
#define EXIT_DUPFD 31

#define MAX_RETRY 15
#define MAX_FILENAME 1024
#define MAX_HOSTNAME 256

static char response(int fd);
static int setup_relay_sock(unsigned int docid,LPS_Pair_t *docattr);
static void connect_relay_socket(unsigned int docid,
				 LPS_Pair_t *docattr,char *filename);
static void connect_remote_pserver(unsigned int docid,
				   LPS_Pair_t *docattr,wchar_t *rm,
				   wchar_t *rp);

int main(int argc,char **argv){
  LPS_Printer_t *printer;
  LPS_Pair_t *docattr;
  wchar_t *rm;
  wchar_t *rp;
  char *sd;
  wchar_t *tmpwcs;
  int localsock;
  int err;
  off_t filesize;
  char filename[MAX_FILENAME];
  GString *command;
  unsigned int docid;
  int docfd;
  LPS_System_t *sys=lps_init(NULL);

  docid=lps_filter_opts(sys,argc,argv,&printer,&docattr);
  // sanity checks
  if((rm=lps_pr_lookup_field(printer,L"rm"))==NULL){
    fprintf(stderr,"Printer Configuration Error: No remote server "
	    "specified\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_NORM);
  } 
  if((rp=lps_pr_lookup_field(printer,L"rp"))==NULL){
    fprintf(stderr,"Printer Configuration Error: No remote queue name\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_NORP);
  }

  // XXX I could potentially provide a default place to stick
  // everything.
  if((sd=lps_chdir_sd(printer,&err))==NULL){
    switch(err){
    case 1:
      fprintf(stderr,"Printer Configuration Error: No spool directory\n");
      lps_fail_doc(docid,docattr);
      exit(EXIT_NOSD);
    case 2:
      fprintf(stderr,"Internal Error: Out of memory.\n");
      lps_fail_doc(docid,docattr);
      exit(EXIT_NOMEM);
    case 3:
      fprintf(stderr,"Printer Configuration Error: Can't cd to spooldir: %s\n",
	      sys_errlist[errno]);
      lps_fail_doc(docid,docattr);
      exit(EXIT_BADSD);
    }
  }

  // if we are not the last document
  if((tmpwcs=lps_pair_lookup(docattr,LPS_AGG_LAST))==NULL || 
     wcscmp(tmpwcs,L"true")==0)
    localsock=setup_relay_sock(docid,docattr);

  // if we are the first document
  if((tmpwcs=lps_pair_lookup(docattr,LPS_AGG_SEQNO))==NULL || 
     wcscmp(tmpwcs,L"1")==0){
    //open a socket to the remote machine
    connect_remote_pserver(docid,docattr,rm,rp);
    strcpy(filename,"dfA");
    strcat(filename,argv[2]+strlen(argv[2])-3);
    gethostname(filename+strlen(filename),MAX_FILENAME-strlen(filename)-1);
  }else{
    /* connect to the previous doc's relay socket and get the
       connection to the remote print server */
    connect_relay_socket(docid,docattr,filename);
    // increment the filename
    filename[2]++;
    if(filename[2]=='[')
      filename[2]='a';
    if(filename[2]=='{'){
      fprintf(stderr,"Internal Error: Too many documents in print job\n");
      lps_fail_doc(docid,docattr);
      exit(EXIT_JOBSIZE);
    }  
  }
  //stdout is now connected to the remote print server

  // send the data through
  /* We have to spool the file to disk because we need the length
     of the file before we send it. */
  if((docfd=open(argv[2],O_CREAT|O_RDWR|O_EXCL,600))==-1){
    fprintf(stderr,"Internal Error: Can't open document spoolfile: %s\n",
	    sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_DOCFILE);
  }
  if(lps_send_data(docfd,0)!=1){
    fprintf(stderr,"Internal Error: Writing data into spoolfile: %s\n", 
	    sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_SPOOLFILE);
  }

  if(dup2(docfd,0)==-1){
    fprintf(stderr,"Internal Error: Can't dup2 docfd: %s\n",
	    sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_DUPDOCFILE);
  }
  close(docfd);
  filesize=lseek(0,0,SEEK_END);
  lseek(0,0,SEEK_SET);
  
  command=g_string_new("\003");
  g_string_sprintfa(command,"%ld %s\n",filesize,filename);
  if(write(0,command->str,command->len)==-1){
    fprintf(stderr,"Internal Error: Could not write datafile command %s\n",
	    sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_DFCOMMAND);
  }
  
  if(response(0)!=0){
    fprintf(stderr,"Remote Error: Remote did not accept data file.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_DATAFILE);
  }

  tmpwcs=lps_pair_lookup(docattr,LPS_AGG_LAST);
  // if we are not the last document
  if(wcscmp(tmpwcs,L"true")){
    // send the fd to the next process
    struct sockaddr addr;
    socklen_t addrlen=sizeof(addr);
    struct iovec iov[1];
    struct msghdr msg;
    struct cmsghdr *fdhdr;
    union {
      struct cmsghdr cm;
      char control[CMSG_SPACE(sizeof(int))];
    } control_un;
    int newsock;

    if((newsock=accept(localsock,&addr,&addrlen))==-1){
      fprintf(stderr,"Internal Error: Could not accept relay: %s\n",
	      sys_errlist[errno]);
      lps_fail_doc(docid,docattr);
      exit(EXIT_RELAYACCEPT);
    }
    iov[0].iov_base=filename;
    iov[0].iov_len=strlen(filename);
    msg.msg_control=control_un.control;
    msg.msg_controllen=sizeof(control_un.control);
    msg.msg_name=NULL;
    msg.msg_namelen=0;
    msg.msg_iov=iov;
    msg.msg_iovlen=1;
    msg.msg_flags=0;
    fdhdr=CMSG_FIRSTHDR(&msg);
    fdhdr->cmsg_len=CMSG_LEN(sizeof(int)); 
    fdhdr->cmsg_level=SOL_SOCKET;
    fdhdr->cmsg_type=SCM_RIGHTS;
    *((int*)CMSG_DATA(fdhdr))=0;
    if(sendmsg(newsock,&msg,0)==-1){
      fprintf(stderr,"Internal Error: Cannot send relay: %s\n",
	      sys_errlist[errno]);
      lps_fail_doc(docid,docattr);
      exit(EXIT_RELAYWRITE);
    }
  }else{
    // send the control file
    GString *controlfile=g_string_new("");
    char tmpstr[MAX_HOSTNAME];
    struct passwd *pwbuf;
    char *str2;
    char letter=filename[2];
    char cur;
    
    filename[2]='A';

    gethostname(tmpstr,MAX_HOSTNAME);
    g_string_sprintfa(controlfile,"H%s\n",tmpstr);
    
    tmpwcs=lps_pair_lookup(docattr,LPS_DOC_UID);
    str2=lps_demote(tmpwcs);
    pwbuf=getpwuid(atoi(str2));
    free(str2);
    g_string_sprintfa(controlfile,"P%s\n",pwbuf->pw_name);
    if((tmpwcs=lps_pair_lookup(docattr,LPS_JOBNAME))!=NULL)
      g_string_sprintfa(controlfile,"J%ls\n",tmpwcs);
    if((tmpwcs=lps_pair_lookup(docattr,LPS_JOBCLASS))!=NULL)
      g_string_sprintfa(controlfile,"C%ls\n",tmpwcs);
    if((tmpwcs=lps_pair_lookup(docattr,LPS_JOBBANNER))!=NULL)
      g_string_sprintfa(controlfile,"L%ls\n",tmpwcs);
    for(cur='A';cur<=letter;cur++){
      if(cur=='['){
	cur='a';
	continue;
      }
      filename[2]=cur;
      g_string_sprintfa(controlfile,"f%s\nU%s\n",filename,filename);
      if((tmpwcs=lps_pair_lookup(docattr,LPS_DOCNAME))!=NULL)
	g_string_sprintfa(controlfile,"L%ls\n",tmpwcs);
    }

    filename[0]='c';
    filename[2]='A';

    g_string_sprintf(command,"\002%d %s\n",controlfile->len,filename);
    if(response(0)!=0){
      fprintf(stderr,"Remote Error: Remote did not accept control file.\n");
      lps_fail_doc(docid,docattr);
      exit(EXIT_DATAFILE);
    }
    write(0,controlfile->str,controlfile->len);
  }

  // exit
  exit(EXIT_OK);
}

char response(int fd){
  char resp;

  if (read(fd, &resp, 1) != 1) 
    return (-1);
  return (resp);
}

int setup_relay_sock(unsigned int docid,LPS_Pair_t *docattr){
  struct sockaddr_un *addr;
  socklen_t len=0;
  GString *str=g_string_new("");
  int localsock;

  if((localsock=socket(PF_LOCAL,SOCK_STREAM,0))==-1){
    fprintf(stderr,"Internal Error: Can't create relay socket: %s\n",
	    sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_RELAYSOCK);
  }

  g_string_sprintfa(str,"/relay.%d",docid);
  len = sizeof(struct sockaddr_un) + str->len + 1;
  if ((addr = malloc(len))==NULL){
    fprintf(stderr,"Internal Error: Out of memory.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_NOMEM);
  }
  
  /* Only need to blank the struct part */
  memset(addr, 0, sizeof(struct sockaddr_un));
  addr->sun_family = AF_LOCAL;
  memcpy(addr->sun_path, str->str, str->len+1);
  g_string_free(str,TRUE);
  
  if(bind(localsock,(struct sockaddr*)addr, SUN_LEN(addr))==-1){
    fprintf(stderr,"Internal Error: Cannot bind socket %s: %s", 
	    str->str,sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_BINDRELAY);
  }

  if(listen(localsock,1)==-1){
    fprintf(stderr,"Internal Error: Cannot listen on socket: %s", 
	    sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_LISTENRELAY);
  }
  return localsock;
}

/* side effect of changing stdout to the remote server */
void connect_relay_socket(unsigned int docid,LPS_Pair_t *docattr,
			  char *filename){
  int relay=socket(PF_LOCAL,SOCK_STREAM,0);
  struct sockaddr_un *addr;
  socklen_t addr_len;
  GString *str=g_string_new(LPS_RELAYSOCK_PATH);
  char *prev;
  
  struct msghdr msg;
  struct msghdr;
  struct iovec iov[1];
  ssize_t len_read;
  union {
    struct cmsghdr cm;
    char control[CMSG_SPACE(sizeof(int))];
  } control_un;
  struct cmsghdr  *fdhdr;
  wchar_t *tmp_wcs;

  if(relay==-1){
    fprintf(stderr,"Internal Error: Can't create socket to accept relay: "
	    "%s\n",sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_RELAYSOCK2);
  }

  if((tmp_wcs=lps_pair_lookup(docattr,L"aggregate_previous"))==NULL){
    fprintf(stderr,"Internal Error: Previous document not specified.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_NOPREV);
  }
    
  if((prev=lps_demote(tmp_wcs))==NULL){
    fprintf(stderr,"Internal Error: Out of memory.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_NOMEM);
  }

  g_string_sprintfa(str,"/relay.%s",prev);
  free(prev);
  addr_len=sizeof(struct sockaddr_un)+str->len+1;
  if ((addr = malloc(addr_len))==NULL){
    fprintf(stderr,"Internal Error: Out of memory.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_NOMEM);
  }
  
  /* Only need to blank the struct part */
  memset(addr, 0, sizeof(struct sockaddr_un));
  addr->sun_family = AF_LOCAL;
  memcpy(addr->sun_path, str->str, str->len+1);
  g_string_free(str,TRUE);

  if(connect(relay,addr,SUN_LEN(addr))==-1){
    fprintf(stderr,"Internal Error: Cannot connect to previous job's "
	    "relay.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_PREVRELAY);
  }

  iov[0].iov_base=&filename;
  iov[0].iov_len=MAX_FILENAME;
  msg.msg_name=NULL;
  msg.msg_namelen=0;
  msg.msg_iov=iov;
  msg.msg_iovlen=1;
  msg.msg_flags=0;
  msg.msg_control=control_un.control;
  msg.msg_controllen=sizeof(control_un.control);

  switch(len_read=recvmsg(relay,&msg,0)){
  case 0:
    fprintf(stderr,"Internal Error: Previous doc's relay close connection "
	    "prematurely.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_HANDOFF);
  case -1:
    fprintf(stderr,"Internal Error: Error reading from relay socket: %s\n",
	    sys_errlist[errno]);
    lps_fail_doc(docid,docattr);
    exit(EXIT_BADRELAY);
  }

  if((fdhdr=CMSG_FIRSTHDR(&msg))==NULL){
    fprintf(stderr,"Internal Error: Relay socket had no ancillary data.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_BADRELAY);
  }

  if(fdhdr->cmsg_level!=SOL_SOCKET){
    fprintf(stderr,"Internal Error: Relay socket ancillary data on wrong "
	    "level.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_BADRELAY);      
  }

  if(fdhdr->cmsg_type!=SCM_RIGHTS){
    fprintf(stderr,"Internal Error: Relay socket ancillary data of wrong "
	    "type.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_BADRELAY);      
  }
  close(relay);
  relay=*((int*)CMSG_DATA(fdhdr));
    
  if(dup2(relay,1)==-1){
    fprintf(stderr,"Internal Error: Relay socket cannot be dup2'd to "
	    "stdout.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_RELAYDUP);
  }
  close(relay);


}

/* side effect of changing stdout to the remote server */
void connect_remote_pserver(unsigned int docid,LPS_Pair_t *docattr,
			    wchar_t *rm,wchar_t *rp){
  struct sockaddr_in sin;
  char *server=lps_demote(rm);
  struct hostent *host;
  char resp;
  int retry;

  if(server==NULL){
    fprintf(stderr,"Internal Error: Out of memory.\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_NOMEM);
  }

  if((host=gethostbyname(server))==NULL){
    fprintf(stderr,"Printer Configuration Error: Can't resolve remote  "
	    "server name\n");
    lps_fail_doc(docid,docattr);
    exit(EXIT_BADRM);
  }
  memset(&sin,0,sizeof(sin));
  sin.sin_port=515;
  free(server);

  // TODO: pick one of the potentially many ipaddrs
  memcpy(&(sin.sin_addr),host->h_addr,host->h_length);

  for(retry=1;retry && retry<MAX_RETRY;){
    int sock;
    if((sock=socket(PF_INET,SOCK_STREAM,0))==-1){
      fprintf(stderr,"Internal Error: Can't create socket to remote print "
	      "server: %s\n",sys_errlist[errno]);
      lps_fail_doc(docid,docattr);
      exit(EXIT_SOCK);
    }

    if(connect(sock,&sin,sizeof(sin))==-1){
      fprintf(stderr,"Internal Error: Can't connect to remote printer: %s\n",
	      sys_errlist[errno]);
      lps_fail_doc(docid,docattr);
      exit(EXIT_NOCONNECT);
    }

    if(dup2(sock,1)==-1){
      fprintf(stderr,"Internal Error: Can't redirect stdout: %s\n", 
	      sys_errlist[errno]);
      lps_fail_doc(docid,docattr);
      exit(EXIT_DUPFD);
    }

    // negotiate the LPR connection
    server=lps_demote(rp);
    do{
      wchar_t *tmpwcs;

      write(1,"\2",1); // 2 is lpr protocol for sending job.
      write(1,server,strlen(server));
      write(1,"\n",1);
      resp=response(1);
      if((resp=response(1))<0 || resp=='\1'){
	// retry the connection
	GString *str=g_string_new("");

	g_string_sprintf(str,"%d",retry++);
	tmpwcs=lps_promote(str->str);
	g_string_free(str,TRUE);
	  
	docattr=lps_pair_update(docattr,L"state",L"retrying_connection");
	docattr=lps_pair_update(docattr,L"queue_retry",tmpwcs);
	lps_commit_doc_attrib(docid,docattr);
	free(tmpwcs);
	retry++;
	break; // goes back to the for loop
      }
      if(resp==0)
	retry=0;
      if(resp>0){
	GString *str=g_string_new("");
	g_string_sprintf(str,"%d",retry++);
	tmpwcs=lps_promote(str->str);
	g_string_free(str,TRUE);
	  
	// queue is full
	sleep(5*60);
	// wait 5 min at least 10 times
	docattr=lps_pair_update(docattr,L"state",L"queue_full");
	docattr=lps_pair_update(docattr,L"queue_retry",tmpwcs);
	lps_commit_doc_attrib(docid,docattr);
	free(tmpwcs);
      }
    }while(resp==0);
  } // end of for(;;)
  if(retry){
    // the connection failed.
    lps_fail_doc(docid,docattr);
    exit(EXIT_BADCONN);
  }
}
